!*************************************************************************
!*                                                                       *
!*     Computation of function and subgradients values of minimum        *
!*     sum of squares clustering problem. Computation with both the      *
!*     whole function and the function using its DC representation are   *
!*     available. (last modified 31.05.2016).                            *
!*                                                                       *
!*************************************************************************
!*
!*     Modules included:
!*
!*     functionmod         !

MODULE functionmod ! Computation of function and (sub)gradients values for clustering software

    USE r_precision, ONLY : prec      ! Precision for reals.
    IMPLICIT NONE

    ! MODULE functionmod includes the following subroutines (S) and functions (F).

    PUBLIC :: &
        auxfunc, &         ! S Computation of aux clustering problem.
        auxfunc1, &        ! S Computation of f1 for aux clustering problem.
        auxfunc2, &        ! S Computation of f2 for aux clustering problem.
        clusterfunc, &     ! S Computation of clustering problem.
        clusterfunc1, &    ! S Computation of f1 for clustering problem.
        clusterfunc2, &    ! S Computation of f2 for clustering problem.
        fgrad, &           ! S Computation of the subgradient of the (aux) clustering problem.
        fgrad1, &          ! S Computation of the gradient of the first component function.
        fgrad2             ! S Computation of the subgradient of the second component function.
!    PRIVATE :: &


CONTAINS

    !=============================================
    ! Computation of aux clustering problem
    !=============================================
    SUBROUTINE auxfunc(x,fval)

        USE param, ONLY : zero, one
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord)
            tnorm          ! The total number of norms computed.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            fval

        ! Local variables
        REAL(KIND=prec) ::  &
            f1,f2,f3
        INTEGER :: &
            i,j

        INTRINSIC ::  DMIN1

        fval=zero
        DO i=1,nrecord
            f1=dminim(i)
            tnorm=tnorm+one
            f2=zero
            DO j=1,mf
                f2=f2+(a(i,j)-x(j))**2
            END DO
            f3=DMIN1(f1,f2)
            fval=fval+f3
        END DO
        RETURN

    END SUBROUTINE auxfunc


    !=============================================
    ! Computation of f_1 aux clustering problem
    !=============================================
    SUBROUTINE auxfunc1(x,fval)

        USE param, ONLY : zero, one
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord)
            tnorm          ! The total number of norms computed.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            fval

        ! Local variables
        REAL(KIND=prec) ::  &
            f1,f2
        INTEGER :: &
            i,j

        INTRINSIC ::  DMIN1

        fval=zero
        DO i=1,nrecord
            f1=dminim(i)
            tnorm=tnorm+one
            f2=zero
            DO j=1,mf
                f2=f2+(a(i,j)-x(j))**2
            END DO
            fval=fval+f1+f2
        END DO
        RETURN

    END SUBROUTINE auxfunc1

    !=============================================
    ! Computation of f_2 aux clustering problem
    !=============================================
    SUBROUTINE auxfunc2(x,fval)

        USE param, ONLY : zero, one
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord)
            tnorm          ! The total number of norms computed.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            fval

        ! Local variables
        REAL(KIND=prec) ::  &
            f1,f2,f3
        INTEGER :: &
            i,j

        INTRINSIC ::  DMIN1

        fval=zero
        DO i=1,nrecord
            f1=dminim(i)
            tnorm=tnorm+one
            f2=zero
            DO j=1,mf
                f2=f2+(a(i,j)-x(j))**2
            END DO
            f3=DMAX1(f1,f2)
            fval=fval+f3
        END DO
        RETURN

    END SUBROUTINE auxfunc2



    !=============================================
    ! Computation of clustering problem
    !=============================================
    SUBROUTINE clusterfunc(x,f)

        USE param, ONLY : zero, one, large
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix.
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nc, &          ! Current number of clusters, loops from 1 to nclust.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord).
            tnorm          ! The total number of norms computed.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            f

        ! Local variables
        REAL(KIND=prec) ::  &
            f1,f2,f3
        INTEGER :: &
            i,j,k

        INTRINSIC ::  DMIN1

        f=zero
        DO i=1,nrecord
            f2=large
            DO k=1,nc
                tnorm=tnorm+one
                f3=zero
                DO j=1,mf
                    f3=f3+(a(i,j)-x(j+(k-1)*mf))**2
                END DO
                f2=DMIN1(f2,f3)
            END DO
            f=f+f2
        END DO

        RETURN

    END SUBROUTINE clusterfunc


    !=============================================
    ! Computation of f_1 for clustering problem
    !=============================================
    SUBROUTINE clusterfunc1(x,f)

        USE param, ONLY : zero, one, large
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix.
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nc, &          ! Current number of clusters, loops from 1 to nclust.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord).
            tnorm          ! The total number of norms computed.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            f

        ! Local variables
        REAL(KIND=prec) ::  &
            f3
        INTEGER :: &
            i,j,k

        INTRINSIC ::  DMIN1

        f=zero
        DO i=1,nrecord
            DO k=1,nc
                tnorm=tnorm+one
                f3=zero
                DO j=1,mf
                    f3=f3+(a(i,j)-x(j+(k-1)*mf))**2
                END DO
                f=f+f3
            END DO
        END DO

        RETURN

    END SUBROUTINE clusterfunc1

    !=============================================
    ! Computation of f_2 for clustering problem
    !=============================================

    SUBROUTINE clusterfunc2(x,f)

        USE param, ONLY : zero, one, large
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix.
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nc, &          ! Current number of clusters, loops from 1 to nclust.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord).
            tnorm          ! The total number of norms computed.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            f

        ! Local variables
        REAL(KIND=prec) ::  &
            f1,f2,f3
        INTEGER :: &
            i,j,k,l

        INTRINSIC ::  DMIN1

        f=zero
        DO i=1,nrecord

            f2=-large
            DO k=1,nc
                !                tnorm=tnorm+one
                f3=zero
                DO l=1,k-1
                    DO j=1,mf
                        f3=f3+(a(i,j)-x(j+(l-1)*mf))**2
                    END DO
                END DO
                DO l=k+1,nc
                    DO j=1,mf
                        f3=f3+(a(i,j)-x(j+(l-1)*mf))**2
                    END DO
                END DO
                f2=MAX(f2,f3)
            END DO

            f=f+f2
        END DO

        RETURN

    END SUBROUTINE clusterfunc2

    !==============================================================
    ! Computation of the gradient of the first component function
    !==============================================================
    SUBROUTINE fgrad(x,grad)

        USE param, ONLY : zero, one, two, large
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix.
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nc, &          ! Current number of clusters, loops from 1 to nclust.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord).
            tnorm, &       ! The total number of norms computed.
            ns, &          ! Switch for auxiliary and real clustering problem.
            m              ! Number of variables in optimization:
                           !   m = mf    if ns = 1,
                           !   m = mf*nc if ns = 2.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            grad(maxdim)

        ! Local variables
        REAL(KIND=prec) ::  &
            f12,f3
        INTEGER :: &
            i,j,k,jmin

        DO i=1,m
            grad(i)=zero
        END DO

        IF (ns == 1) THEN
            DO i=1,nrecord
                tnorm=tnorm+one
                f3=zero
                DO j=1,mf
                    f3=f3+(a(i,j)-x(j))**2
                END DO
                IF (f3 < dminim(i)) THEN
                    DO j=1,m
                        grad(j)=grad(j)+two*(x(j)-a(i,j))
                    END DO
                END IF
            END DO

        ELSE
            DO i=1,nrecord
                f12=large
                DO k=1,nc
                    tnorm=tnorm+one
                    f3=zero
                    DO j=1,mf
                        f3=f3+(a(i,j)-x(j+(k-1)*mf))**2
                    END DO
                    IF (f12 > f3) THEN
                        f12=f3
                        jmin=k
                    END IF
                END DO
                DO j=1,mf
                    grad(j+(jmin-1)*mf)=grad(j+(jmin-1)*mf)+two*(x(j+(jmin-1)*mf)-a(i,j))
                END DO
            END DO
        END IF

    END SUBROUTINE fgrad

    !==============================================================
    ! Computation of the gradient of the first component function
    !==============================================================
    SUBROUTINE fgrad1(x1,grad1)

        USE param, ONLY : zero, two
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix.
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nc, &          ! Current number of clusters, loops from 1 to nclust.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord).
            ns, &          ! Switch for auxiliary and real clustering problem.
            m              ! Number of variables in optimization:
                           !   m = mf    if ns = 1,
                           !   m = mf*nc if ns = 2.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x1(maxdim), &
            grad1(maxdim)

        ! Local variables
        INTEGER :: &
            i,j,k

        DO i=1,m
            grad1(i)=zero
        END DO

        IF (ns==1) THEN
            DO i=1,nrecord
                DO j=1,m
                    grad1(j)=grad1(j)+two*(x1(j)-a(i,j))
                END DO
            END DO
        END IF

        IF (ns==2) THEN
            DO i=1,nrecord
                DO j=1,nc
                    DO k=1,mf
                        grad1(k+(j-1)*mf)=grad1(k+(j-1)*mf)+two*(x1(k+(j-1)*mf)-a(i,k))
                    END DO
                END DO
            END DO
        END IF

        RETURN

    END SUBROUTINE fgrad1


    !==============================================================
    ! Computation of the gradient of the second component function
    !==============================================================
    SUBROUTINE fgrad2(x,grad2)

        USE param, ONLY : zero, one, two
        USE initclust, ONLY : &
            maxdim, &      ! Maximum number of variables in optimization,
                           !   maxdim >= nclust * nft.
            a, &           ! Data matrix.
            mf, &          ! Number of used features:
                           !   mf = nft, when no classes,
                           !   mf = nft - 1, when classes.
            nc, &          ! Current number of clusters, loops from 1 to nclust.
            nrecord, &     ! Number of reports in data.
            dminim, &      ! dminim(nrecord).
            tnorm, &       ! The total number of norms computed.
            ns, &          ! Switch for auxiliary and real clustering problem.
            m              ! Number of variables in optimization:
                           !   m = mf    if ns = 1,
                           !   m = mf*nc if ns = 2.

        IMPLICIT NONE

        ! Subroutine arguments
        REAL(KIND=prec) ::  &
            x(maxdim), &
            grad2(maxdim)

        ! Local variables
        REAL(KIND=prec) ::  &
            d2(nc), &
            d3,d4
        INTEGER :: &
            i,j,k,jindex

        DO i=1,m
            grad2(i)=zero
        END DO

        IF (ns==1) THEN
            DO i=1,nrecord
                tnorm=tnorm+one
                d3=zero
                DO j=1,mf
                    d3=d3+(a(i,j)-x(j))**2
                END DO
                IF (d3 > dminim(i)) THEN
                    DO j=1,m
                        grad2(j)=grad2(j)+two*(x(j)-a(i,j))
                    END DO
                END IF
            END DO
        END IF

        IF (ns == 2) THEN
            DO i=1,nrecord
                DO j=1,nc
                    tnorm=tnorm+one
                    d2(j)=zero
                    DO k=1,mf
                        d2(j)=d2(j)+(a(i,k)-x(k+(j-1)*mf))**2
                    END DO
                END DO
                d3=zero
                DO j=1,nc
                    d4=zero
                    DO k=1,nc
                        IF (k.ne.j) THEN
                            d4=d4+d2(k)
                        END IF
                    END DO
                    IF (d3<d4) THEN
                        d3=d4
                        jindex=j
                    END IF
                END DO
                DO j=1,nc
                    IF (j.ne.jindex) THEN
                        DO k=1,mf
                            grad2(k+(j-1)*mf)=grad2(k+(j-1)*mf)+two*(x(k+(j-1)*mf)-a(i,k))
                        END DO
                    END IF
                END DO
            END DO
        END IF

        RETURN

    END SUBROUTINE fgrad2

END MODULE functionmod
